# -*- coding: utf-8 -*-

"""Minimization step in LAMMPS"""

import logging

from seamm_util import Q_, units_class
from seamm_util.printing import FormattedText as __
import lammps_step

logger = logging.getLogger(__name__)


class Minimization(lammps_step.Energy):
    def __init__(self, flowchart=None, title="Minimization", extension=None):
        """Initialize the node"""

        logger.debug("Creating Minimization {}".format(self))

        super().__init__(flowchart=flowchart, title=title, extension=extension)

        self.description = "Minimization step in LAMMPS"

        self._calculation = "minimization"
        self._model = None
        self._metadata = lammps_step.metadata

        self.convergence = "normal"
        self.etol_method = "is"
        self.etol = 1.0e-06
        self.etol_variable = ""
        self.ftol_method = "is"
        self.ftol = Q_(0.1, "kcal/mol/Å")
        self.ftol_variable = ""
        self.maxiters_method = "is"
        self.maxiters = 10000
        self.maxiters_variable = ""
        self.maxevals_method = "is"
        self.maxevals = 30000
        self.maxevals_variable = ""

    def description_text(self, P=None):
        """Create the text description of what this step will do."""

        # if not P:
        #     P = self.parameters.values_to_dict()

        text = "Structure optimization calculation."

        return self.header + "\n" + __(text, indent=4 * " ").__str__()

    def get_input(self, extras=None):
        """Get the input for a minimization in LAMMPS"""

        P = self.parameters.values_to_dict()

        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        self.description = []
        self.description.append(__(self.description_text(PP), **PP, indent=3 * " "))

        system_db = self.get_variable("_system_db")
        configuration = system_db.system.configuration
        n_atoms = configuration.n_atoms
        nDOF = 3 * n_atoms

        lines = []

        etol = 0.0
        ftol = None
        maxiters = None
        maxevals = None
        if self.convergence == "normal":
            etol = 0.0
            ftol = 0.1
            maxiters = 3 * nDOF
            maxevals = 5 * maxiters
        elif self.convergence == "tight":
            etol = 0.0
            ftol = 0.01
            maxiters = 2 * 3 * nDOF
            maxevals = 5 * maxiters
        elif self.convergence == "loose":
            etol = 0.0
            ftol = 1
            maxiters = nDOF
            maxevals = 5 * maxiters
        elif self.convergence == "crude":
            etol = 0.0
            ftol = 10
            maxiters = int(nDOF / 10)
            maxevals = 2 * maxiters
        elif "energy" in self.convergence:
            # energy tolerance
            if self.etol_method == "is":
                etol = self.etol
            else:
                raise RuntimeError("Variable handling not implemented for etol")
        elif "forces" in self.convergence:
            # force tolerance
            if self.ftol_method == "is":
                ftol = self.ftol.to("kcal/mol/Å").magnitude
            else:
                raise RuntimeError("Variable handling not implemented for ftol")
        else:
            raise RuntimeError("Variable handling not implemented for convergence")

        # maximum number of iterations
        if maxiters is None:
            if self.maxiters_method == "default":
                maxiters = 3 * nDOF
            elif self.maxiters_method == "is":
                maxiters = self.maxiters
            else:
                raise RuntimeError("Variable handling not implemented for maxiters")

        # maximum number of energy evaluations
        if maxevals is None:
            if self.maxevals_method == "default":
                maxevals = 3 * nDOF
            elif self.maxevals_method == "is":
                maxevals = self.maxevals
            else:
                raise RuntimeError("Variable handling not implemented for maxevals")

        thermo_properties = (
            "press etotal ke pe ebond " "eangle edihed eimp evdwl etail ecoul elong"
        )

        lines.append("")
        lines.append("#     Minimization")
        lines.append("")
        lines.append("thermo_style        custom {}".format(thermo_properties))
        lines.append("thermo              {}".format(100))
        lines.append(
            "minimize            {} {} {} {}".format(etol, ftol, maxiters, maxevals)
        )

        return {
            "script": lines,
            "postscript": None,
            "use python": False,
        }



# -*- coding: utf-8 -*-

"""NVT (canonical) dynamics in LAMMPS"""

import json

import lammps_step
import logging
import seamm
from seamm_util import units_class, Q_
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __
import random

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("lammps")

thermostat_metadata = {
    "Nose-Hoover": {
        "documentation": "http://lammps.sandia.gov/doc/fix_nvt.html",
        "references": ["Shinoda", "Tuckerman"],
    },
    "Berendsen": {
        "documentation": "http://lammps.sandia.gov/doc/fix_temp_berendsen.html",
        "references": ["Berendsen"],
    },
    "canonical sampling, velocity rescaling (csvr)": {
        "documentation": "http://lammps.sandia.gov/doc/fix_temp_csvr.html",
        "references": ["Bussi1"],
    },
    "canonical sampling, langevin dynamics (csld)": {
        "documentation": "http://lammps.sandia.gov/doc/fix_temp_csvr.html",
        "references": ["Bussi2"],
    },
    "velocity rescaling": {
        "documentation": "http://lammps.sandia.gov/doc/fix_temp_rescale.html",
        "references": [],
    },
    "Langevin": {
        "documentation": "http://lammps.sandia.gov/doc/fix_langevin.html",
        "references": ["Schneider", "Dunweg"],
    },
}


class NVT(lammps_step.NVE):
    def __init__(
        self,
        flowchart=None,
        title="NVT dynamics",
        extension=None,
        logger=logger,
    ):
        """Initialize the node"""

        logger.debug("Creating NVT {}".format(self))

        super().__init__(
            flowchart=flowchart,
            title=title,
            extension=extension,
            logger=logger,
        )

        self.logger.debug("NVT after super init, {}".format(self))

        self.description = "NVT dynamics step in LAMMPS"

        self.logger.debug("NVT.init() creating NVT_Parameters object")

        self._calculation = "nvt"
        self._model = None
        self._metadata = lammps_step.metadata
        self.parameters = lammps_step.NVT_Parameters()

        self.logger.debug("NVT.init() completed")

    def description_text(self, P=None):
        """Return a short description of this step.

        Return a nicely formatted string describing what this step will
        do.

        Keyword arguments:
            P: a dictionary of parameter values, which may be variables
                or final values. If None, then the parameters values will
                be used as is.
        """

        if not P:
            P = self.parameters.values_to_dict()

        # What will we do?

        if P["T0"] == P["T1"]:
            text = "{time} of canonical (NVT) dynamics at {T0} "
        else:
            text = (
                "{time} of canonical (NVT) dynamics starting "
                " at {T0}, going to {T1}, "
            )
        if P["thermostat"] == "Nose-Hoover":
            text += "using a Nose-Hoover thermostat."
            if P["Tchain"] != "3":
                if P["Tloop"] != "1":
                    text += (
                        " The thermostat will use a chain of {Tchain} "
                        "thermostats with {Tloop} subcycles and a "
                    )
                else:
                    text += (
                        " The thermostat will use a chain of {Tchain} "
                        "thermostats and a "
                    )
            elif P["Tloop"] != "1":
                text += " The thermostat will use {Tloop} subcycles and a "
            else:
                text += " The thermostat will use a "
            text += "drag factor of {drag}."
        elif P["thermostat"] == "Berendsen":
            text += "using a Berendsen thermostat with a damping time " "of {Tdamp}"
        elif "csvr" in P["thermostat"]:
            text += (
                "using a canonical sampling thermostat using velocity "
                "rescaling (CSVR) with a damping time of {Tdamp} and "
                "a {seed}."
            )
        elif "csld" in P["thermostat"]:
            text += (
                "using a canonical sampling thermostat using Langevin "
                "dynamics (CSLD) with a damping time of {Tdamp} and "
                "a {seed}."
            )
        elif P["thermostat"] == "velocity rescaling":
            text += (
                "using velocity rescaling every {frequency} with a "
                "temperature window of {window}."
            )
            if P["fraction"] != 1.0:
                text += (
                    " The velocities will only be scaled a fraction "
                    "({fraction}) of the amount needed to fully correct "
                    "the temperature."
                )
        elif P["thermostat"] == "Langevin":
            text += (
                "using a Langevin thermostat with a damping time "
                "of {Tdamp} and a {seed}"
            )
        else:
            text += "using the thermostat given by {thermostat}"

        return self.header + "\n" + __(text, **P, indent=4 * " ").__str__()

    def describe(self, indent="", json_dict=None):
        """Write out information about what this node will do
        If json_dict is passed in, add information to that dictionary
        so that it can be written out by the controller as appropriate.
        """

        # Can't call super() because it will print too much
        self.visited = True
        job.job("\n" + self.indent + self.header)
        next_node = self.next()

        # Local copies of variables in a dictionary

        P = self.parameters.values_to_dict()
        text = self.description_text(P)
        job.job(__(text, indent=self.indent + "    ", **P))

        return next_node

    def get_input(self, extras=None):
        """Get the input for an NVT dynamics run in LAMMPS"""

        self.description = []

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        # Fix variables with special cases
        timestep, P["timestep"] = self.timestep(P["timestep"])

        if P["seed"] == "random":
            # Apparently the seed must be no larger than 900,000,000
            P["seed"] = int(random.random() * 899999999)

        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        self.description.append(
            __(self.description_text(PP), **PP, indent=3 * " ").__str__()
        )

        time = lammps_step.to_lammps_units(P["time"], quantity="time")
        nsteps = round(time / timestep)

        T0 = lammps_step.to_lammps_units(P["T0"], quantity="temperature")
        T1 = lammps_step.to_lammps_units(P["T1"], quantity="temperature")
        Tdamp = lammps_step.to_lammps_units(P["Tdamp"], quantity="time")

        thermo_properties = (
            "time temp press etotal ke pe ebond "
            "eangle edihed eimp evdwl etail ecoul elong"
        )
        properties = "v_time v_temp v_press v_etotal v_ke v_pe v_epair"
        title2 = "tstep t T P Etot Eke Epe Epair"

        lines = []
        lines.append("")
        lines.append("#     NVT dynamics")
        lines.append("")
        lines.append("reset_timestep      0")
        lines.append("timestep            {}".format(timestep))
        lines.append("thermo_style        custom {}".format(thermo_properties))
        lines.append("thermo              {}".format(int(nsteps / 100)))

        ncomputes = 0
        ndumps = 0
        nfixes = 0
        if P["thermostat"] == "Nose-Hoover":
            Tchain = P["Tchain"]
            Tloop = P["Tloop"]
            drag = P["drag"]
            nfixes += 1
            lines.append(
                "fix                 {} all nvt ".format(nfixes)
                + "temp {} {} {} ".format(T0, T1, Tdamp)
                + "tchain {} ".format(Tchain)
                + "tloop {} ".format(Tloop)
                + "drag {}".format(drag)
            )
        elif P["thermostat"] == "Berendsen":
            nfixes += 1
            lines.append(
                "fix                 {} ".format(nfixes)
                + "all temp/berendsen "
                + " {} {} {}".format(T0, T1, Tdamp)
            )
            nfixes += 1
            lines.append("fix                 {} ".format(nfixes) + "all nve")
        elif "csvr" in P["thermostat"]:
            seed = P["seed"]
            nfixes += 1
            lines.append(
                "fix                 {} ".format(nfixes)
                + "all temp/csvr "
                + " {} {} {} {}".format(T0, T1, Tdamp, seed)
            )
            nfixes += 1
            lines.append("fix                 {} ".format(nfixes) + "all nve")
        elif "csld" in P["thermostat"]:
            seed = P["seed"]
            nfixes += 1
            lines.append(
                "fix                 {} ".format(nfixes)
                + "all temp/csld "
                + " {} {} {} {}".format(T0, T1, Tdamp, seed)
            )
            nfixes += 1
            lines.append("fix                 {} ".format(nfixes) + "all nve")
        elif P["thermostat"] == "velocity rescaling":
            frequency = lammps_step.to_lammps_units(P["frequency"], quantity="time")

            nevery = max(1, round(nsteps / (frequency / timestep)))
            window = lammps_step.to_lammps_units(P["window"], quantity="temperature")
            fraction = P["fraction"]
            nfixes += 1
            lines.append(
                "fix                 {} ".format(nfixes)
                + "all temp/rescale "
                + "{} {} {} {} {}".format(nevery, T0, T1, window, fraction)
            )
            nfixes += 1
            lines.append("fix                 {} ".format(nfixes) + "all nve")
        elif P["thermostat"] == "Langevin":
            seed = P["seed"]
            nfixes += 1
            lines.append(
                "fix                 {} ".format(nfixes)
                + "all langevin "
                + "{} {} {} {} ".format(T0, T1, Tdamp, seed)
            )
            nfixes += 1
            lines.append("fix                 {} ".format(nfixes) + "all nve")
        else:
            raise RuntimeError(
                "Don't recognize temperature control " + "'{}'".format(P["thermostat"])
            )

        # Add the citation for the thermostat
        for citation in thermostat_metadata[P["thermostat"]]["references"]:
            self.references.cite(
                raw=self._bibliography[citation],
                alias=citation,
                module="lammps_step",
                level=1,
                note="Citation for thermostat.",
            )
        # For the heat flux, if requested, we need extra input
        if P["heat flux"] != "never":
            # Unit conversion factor
            if lammps_step.get_lammps_unit_system() == "metal":
                factor = Q_("eV/Å^2/ps")
            else:
                factor = (
                    Q_("kcal/Å^2/fs/mol") / Q_("kcal/mol") * Q_("kcal/mol").to("kJ")
                )
            factor = factor.m_as("W/m^2")
            lines.append(
                f"""
compute             KE all ke/atom
compute             PE all pe/atom

#          centroid doesn't work with kspace, so split into pair and non-pair parts

compute             S_p all stress/atom NULL pair kspace
compute             S_b all centroid/stress/atom NULL bond angle dihedral improper
compute             flux_p all heat/flux KE PE S_p
compute             flux_b all heat/flux KE PE S_b

#          Conversion from kcal/Å^2/fs/mol to W/m^2")

variable            factor equal {factor}
variable            Jx equal v_factor*(c_flux_p[1]+c_flux_b[1])/vol
variable            Jy equal v_factor*(c_flux_p[2]+c_flux_b[2])/vol
variable            Jz equal v_factor*(c_flux_p[3]+c_flux_b[3])/vol
"""
            )

        # summary output written 10 times during run so we can see progress
        nevery = 10
        nfreq = int(nsteps / 10)
        nrepeat = int(nfreq / nevery)
        nfreq = nevery * nrepeat
        nfixes += 1
        filename = f"@{self._id[-1]}+nvt_summary.trj"
        lines.append(
            f"fix                 {nfixes} all ave/time {nevery} 1 {nfreq} &\n"
            f"                       {properties} &\n"
            "                       off 2 &\n"
            f"                       file {filename}"
        )
        # instantaneous output written for averaging
        if P["sampling"] == "none":
            self.description.append(
                __(
                    "The run will be {nsteps:n} steps of dynamics.",
                    nsteps=nsteps,
                    indent=7 * " ",
                )
            )
        else:
            sampling = lammps_step.to_lammps_units(P["sampling"], quantity="time")
            nevery = max(1, round(sampling / timestep))
            nrepeat = 1
            nfreq = nevery * nrepeat
            nfixes += 1

            dt = (nevery * P["timestep"]).to_compact()
            if T0 == T1:
                text = json.dumps(
                    {
                        "code": "LAMMPS",
                        "type": "NVT",
                        "dt": dt.magnitude,
                        "tunits": str(dt.u),
                        "nsteps": nsteps // nevery,
                        "T": T0,
                        "Tunits": "K",
                    },
                    separators=(",", ":"),
                )
            else:
                text = json.dumps(
                    {
                        "code": "LAMMPS",
                        "type": "NVT",
                        "dt": dt.magnitude,
                        "units": str(dt.u),
                        "nsteps": nsteps // nevery,
                        "T0": T0,
                        "T1": T1,
                        "Tunits": "K",
                    },
                    separators=(",", ":"),
                )
            title1 = "!MolSSI trajectory 2.0 " + text
            filename = f"@{self._id[-1]}+nvt_state.trj"
            lines.append(
                f"fix                 {nfixes} all ave/time {nevery} 1 {nfreq} &\n"
                f"                       {properties} &\n"
                "                       off 2 &\n"
                f"                       title1 '{title1}' &\n"
                f"                       title2 '{title2}' &\n"
                f"                       file {filename}"
            )
            self.description.append(
                __(
                    (
                        "The run will be {nsteps:,d} steps of dynamics "
                        "sampled every {nevery:n} steps."
                    ),
                    nsteps=nsteps,
                    nevery=nevery,
                    indent=7 * " ",
                )
            )

        # Handle trajectories
        tmp, ncomputes, ndumps, nfixes = self.trajectory_input(
            P, timestep, nsteps, ncomputes, ndumps, nfixes
        )
        lines.extend(tmp)

        if extras is not None and "shake" in extras:
            nfixes += 1
            lines.append(extras["shake"].format(nfixes))

        lines.append("")
        lines.append("run                 {}".format(nsteps))
        lines.append("")

        for i in range(1, ncomputes + 1):
            lines.append(f"uncompute           {i}")
        for i in range(1, ndumps + 1):
            lines.append(f"undump              {i}")
        for i in range(1, nfixes + 1):
            lines.append(f"unfix               {i}")
        lines.append("")

        return {
            "script": lines,
            "postscript": None,
            "use python": False,
        }



# -*- coding: utf-8 -*-

"""A single-point energy in LAMMPS"""

import logging

from tabulate import tabulate

import lammps_step
import seamm
from seamm_util import units_class
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

logger = logging.getLogger(__name__)
printer = printing.getPrinter("lammps")


class Energy(seamm.Node):
    """Handle a singlepoint energy calculation in LAMMPS"""

    def __init__(
        self,
        flowchart=None,
        title="Energy",
        extension=None,
        logger=logger,
    ):
        """Initialize the node"""

        logger.debug("Creating Energy {}".format(self))

        super().__init__(
            flowchart=flowchart,
            title=title,
            extension=extension,
            logger=logger,
        )

        self._calculation = "energy"
        self._model = None
        self._metadata = lammps_step.metadata
        self.parameters = lammps_step.EnergyParameters()

        self.description = "A single point energy calculation"

    @property
    def header(self):
        """A printable header for this section of output"""
        return "Step {}: {}".format(".".join(str(e) for e in self._id), self.title)

    @property
    def version(self):
        """The semantic version of this module."""
        return lammps_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return lammps_step.__git_revision__

    def description_text(self, P=None):
        """Create the text description of what this step will do."""

        # if not P:
        #     P = self.parameters.values_to_dict()

        text = "Single-point energy calculation."

        return self.header + "\n" + __(text, indent=4 * " ").__str__()

    def get_input(self, extras=None):
        """Get the input for an energy calculation for LAMMPS"""

        P = self.parameters.values_to_dict()

        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        self.description = []
        self.description.append(__(self.description_text(PP), **PP, indent=3 * " "))

        lines = []

        lines.append("")
        lines.append("#     single-point energy")
        lines.append("")
        lines.append("run                 0")

        return {
            "script": lines,
            "postscript": None,
            "use python": False,
        }

    def analyze(self, indent="", data={}, table=None, **kwargs):
        """Parse the output and generating the text output and store the
        data in variables for other stages to access
        """
        if table is not None:
            text = ""
            tmp = tabulate(
                table,
                headers="keys",
                tablefmt="simple",
                disable_numparse=True,
                colalign=(
                    "center",
                    "decimal",
                    "center",
                    "decimal",
                    "left",
                    "decimal",
                    "decimal",
                    "decimal",
                ),
            )
            length = len(tmp.splitlines()[0])
            text += "\n"
            text += "Properties".center(length)
            text += "\n"
            text += tmp
            text += "\n"

            printer.normal(__(text, indent=8 * " ", wrap=False, dedent=False))

            have_warning = False
            for value in table["Property"]:
                if len(value) > 0 and value[-1] == "*":
                    have_warning = True
                    break

            have_acf_warning = False
            for value in table["tau"]:
                if len(value) > 0 and value[-1] == "^":
                    have_warning = True
                    break

            if have_warning or have_acf_warning:
                printer.normal("\n")
            if have_warning:
                printer.normal(
                    __(
                        "          * this property has less than 100 independent "
                        "samples, so may not be accurate.",
                        wrap=False,
                        dedent=False,
                    )
                )

            if have_acf_warning:
                printer.normal(
                    __(
                        "          ^ there are not enough samples after "
                        "equilibration to plot the ACF.",
                        wrap=False,
                        dedent=False,
                    )
                )

        # Get the configuration
        _, configuration = self.get_system_configuration(None)

        # Need to set the model for properties.
        # See what type of forcefield we have and handle it
        ff = self.get_variable("_forcefield")
        if ff == "OpenKIM":
            self._model = "OpenKIM/" + self.get_variable("_OpenKIM_Potential")
        else:
            # Valence forcefield...
            self._model = ff.current_forcefield

        # Put any requested results into variables or tables
        self.store_results(
            configuration=configuration,
            data=data,
            create_tables=self.parameters["create tables"].get(),
        )



# -*- coding: utf-8 -*-

"""A custom script for LAMMPS"""

import seamm
import logging

import lammps_step
from seamm_util.printing import FormattedText as __

logger = logging.getLogger(__name__)


class Custom(seamm.Node):
    def __init__(self, flowchart=None, title="Custom", extension=None):
        """Initialize the node"""

        logger.debug("Creating Custom step {}".format(self))

        super().__init__(flowchart=flowchart, title=title, extension=extension)

        self._calculation = "energy"
        self._model = None
        self._metadata = lammps_step.metadata
        self.parameters = lammps_step.CustomParameters()
        self.description = "A custom script for LAMMPS"

    @property
    def header(self):
        """A printable header for this section of output"""
        return "Step {}: {}".format(".".join(str(e) for e in self._id), self.title)

    @property
    def version(self):
        """The semantic version of this module."""
        return lammps_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return lammps_step.__git_revision__

    def description_text(self, P=None):
        """Create the text description of what this step will do."""

        if not P:
            P = self.parameters.values_to_dict()

        text = "Custom script for LAMMPS\n\n"
        lines = P["script"].splitlines()
        if len(lines) > 5:
            text += "\n".join(lines[0:4]) + "\n...\n"
        else:
            text += P["script"]

        return self.header + "\n" + __(text, indent=4 * " ").__str__()

    def get_input(self, extras=None):
        """Get the custom input for LAMMPS"""

        P = self.parameters.values_to_dict()

        self.description = []
        self.description.append(__(self.description_text(P), **P, indent=3 * " "))

        lines = []

        lines.append("")
        lines.append("#     custom scripting")
        lines.append("")
        lines.append(P["script"])

        return {
            "script": lines,
            "postscript": None,
            "use python": False,
        }



# -*- coding: utf-8 -*-

"""Installer for the LAMMPS plug-in.

This handles any further installation needed after installing the Python
package `lammps-step`.
"""

import logging
import pkg_resources
from pathlib import Path
import shutil
import subprocess

import seamm_installer

logger = logging.getLogger(__name__)


class Installer(seamm_installer.InstallerBase):
    """Handle further installation needed after installing lammps-step.

    The Python package `lammps-step` should already be installed, using `pip`,
    `conda`, or similar. This plug-in-specific installer then checks for the
    LAMMPS executables, installing them if need be, and registers their
    location in seamm.ini.

    .. note:: The LAMMPS step and this assume that the LAMMPS
       executables, if present, are in the same directory. In
       addition, for the parallel version using MPI it is assumed that
       `mpiexec` is either in that same directory, or the correct
       `mpiexec` can be found in the PATH.

    There are a number of ways to determine which are the correct LAMMPS
    executables to use. The aim of this installer is to help the user locate
    the executables. There are a number of possibilities

    #. The correct executables are already available.

        #. If they are already registered in `seamm.ini` there is nothing else
           to do.

        #. They may be in the current path, in which case they need to be added
           to `seamm.ini`.

        #. If a module system is in use, a module may need to be loaded to give
           access to LAMMPS.

        #. They cannot be found automatically, so the user needs to locate the
           executables for the installer.

    #. LAMMPS is not installed on the machine. In this case they can be
       installed in a Conda environment. There is one choice

        #. They can be installed in a separate environment, `seamm-lammps` by
           default.
    """

    def __init__(self, logger=logger):
        # Call the base class initialization, which sets up the commandline
        # parser, amongst other things.
        super().__init__(logger=logger)

        logger.debug("Initializing the LAMMPS installer object.")

        self.section = "lammps-step"
        self.path_name = "lammps-path"
        self.executables = ["lmp_serial", "lmp_mpi"]
        self.resource_path = Path(pkg_resources.resource_filename(__name__, "data/"))
        # What Conda environment is the default?
        data = self.configuration.get_values(self.section)
        if "conda-environment" in data and data["conda-environment"] != "":
            self.environment = data["conda-environment"]
        else:
            self.environment = "seamm-lammps"

        # The environment.yaml file for Conda installations.
        path = Path(pkg_resources.resource_filename(__name__, "data/"))
        logger.debug(f"data directory: {path}")
        self.environment_file = path / "seamm-lammps.yml"

    def check(self):
        """Check the installation and fix errors if requested.

        If the option `yes` is present and True, this method will attempt to
        correct any errors in the configuration file. Use `--yes` on the
        command line to enable this.

        The information in the configuration file is:

            installation
                How LAMMPS is installed. One of `path`, `modules` or `conda`
            conda-environment
                The Conda environment if and only if `installation` = `conda`
            modules
                The environment modules if `installation` = `modules`
            lammps-path
                The path where the LAMMPS executables are. Automatically
                defined if `installation` is `conda` or `modules`, but given
                by the user is it is `path`.

        Returns
        -------
        bool
            True if everything is OK, False otherwise. If `yes` is given as an
            option, the return value is after fixing the configuration.
        """
        self.logger.debug("Entering check method.")
        if not self.configuration.section_exists(self.section):
            if self.options.yes or self.ask_yes_no(
                "There is no section for the LAMMPS step in the configuration "
                f" file ({self.configuration.path}).\nAdd one?",
                default="yes",
            ):
                self.check_configuration_file()
                print(
                    f"Added the {self.section} section to the configuration file "
                    f"{self.configuration.path}"
                )

        # Get the values from the configuration
        data = self.configuration.get_values(self.section)

        # Save the initial values, if any, of the key configuration variables
        if "lammps-path" in data and data["lammps-path"] != "":
            path = Path(data["lammps-path"]).expanduser().resolve()
            initial_lammps_path = path
        else:
            initial_lammps_path = None
        if "installation" in data and data["installation"] != "":
            initial_installation = data["installation"]
        else:
            initial_installation = None
        if "conda-environment" in data and data["conda-environment"] != "":
            initial_conda_environment = data["conda-environment"]
        else:
            initial_conda_environment = None
        if "modules" in data and data["modules"] != "":
            initial_modules = data["modules"]
        else:
            initial_modules = None

        # Is there a valid lammps-path?
        self.logger.debug(
            "Checking for executables in the initial lammps-path "
            f"{initial_lammps_path}."
        )
        if initial_lammps_path is None or not self.have_executables(
            initial_lammps_path
        ):
            lammps_path = None
        else:
            lammps_path = initial_lammps_path
        self.logger.debug(f"initial-lammps-path = {initial_lammps_path}.")

        # Is there an installation indicated?
        if initial_installation in ("path", "conda", "modules"):
            installation = initial_installation
        else:
            installation = None
        self.logger.debug(f"initial-installation = {initial_installation}.")

        if installation == "conda":
            # Is there a conda environment?
            conda_environment = None
            if initial_conda_environment is None or not self.conda.exists(
                initial_conda_environment
            ):
                if lammps_path is not None:
                    # see if this path corresponds to a Conda environment
                    for tmp in self.conda.environments:
                        tmp_path = self.conda.path(tmp) / "bin"
                        if tmp_path == lammps_path:
                            conda_environment = tmp
                            break
                    if conda_environment is not None:
                        if self.options.yes or self.ask_yes_no(
                            "The Conda environment in the config file "
                            "is not correct.\n"
                            f"It should be {conda_environment}. Fix?",
                            default="yes",
                        ):
                            self.configuration.set_value(
                                self.section, "installation", "conda"
                            )
                            self.configuration.set_value(
                                self.section, "conda-environment", conda_environment
                            )
                            self.configuration.set_value(self.section, "modules", "")
                            self.configuration.save()
                            print(
                                "Corrected the conda environment to "
                                f"{conda_environment}"
                            )
            else:
                # Have a Conda environment!
                conda_path = self.conda.path(initial_conda_environment) / "bin"
                self.logger.debug(
                    f"Checking for executables in conda-path: {conda_path}."
                )
                if self.have_executables(conda_path):
                    # All is good!
                    conda_environment = initial_conda_environment
                    if lammps_path is None:
                        if self.options.yes or self.ask_yes_no(
                            "The lammps-path in the config file is not set,"
                            f"but the Conda environment {conda_environment} "
                            "is.\nFix the lammps-path?",
                            default="yes",
                        ):
                            lammps_path = conda_path
                            self.configuration.set_value(
                                self.section, "lammps-path", lammps_path
                            )
                            self.configuration.set_value(self.section, "modules", "")
                            self.configuration.save()
                            print(f"Set the lammps-path to {conda_path}")
                    elif lammps_path != conda_path:
                        if self.options.yes or self.ask_yes_no(
                            f"The lammps-path in the config file {lammps_path}"
                            "is different from that for  the Conda "
                            f"environment {conda_environment} is.\n"
                            "Use the path from the Conda environment?",
                            default="yes",
                        ):
                            lammps_path = conda_path
                            self.configuration.set_value(
                                self.section, "lammps-path", lammps_path
                            )
                            self.configuration.set_value(self.section, "modules", "")
                            self.configuration.save()
                            print(f"Changed the lammps-path to {conda_path}")
                    else:
                        # Everything is fine!
                        pass
        if installation == "modules":
            print(f"Can't check the actual modules {initial_modules} yet")
            if initial_conda_environment is not None:
                if self.options.yes or self.ask_yes_no(
                    "A Conda environment is given: "
                    f"{initial_conda_environment}.\n"
                    "A Conda environment should not be used when using "
                    "modules. Remove it from the configuration?",
                    default="yes",
                ):
                    self.configuration.set_value(self.section, "conda-environment", "")
                    self.configuration.save()
                    print(
                        "Using modules, so removed the conda-environment from "
                        "the configuration"
                    )
        else:
            if lammps_path is None:
                # No path or executables in the path!
                environments = self.conda.environments
                if self.environment in environments:
                    # Make sure it is first!
                    environments.remove(self.environment)
                    environments.insert(0, self.environment)
                for tmp in environments:
                    tmp_path = self.conda.path(tmp) / "bin"
                    if self.have_executables(tmp_path):
                        if self.options.yes or self.ask_yes_no(
                            "There are no valid executables in the lammps-path"
                            " in the config file, but there are in the Conda "
                            f"environment {tmp}.\n"
                            "Use them?",
                            default="yes",
                        ):
                            conda_environment = tmp
                            lammps_path = tmp_path
                            self.configuration.set_value(
                                self.section, "lammps-path", lammps_path
                            )
                            self.configuration.set_value(
                                self.section, "installation", "conda"
                            )
                            self.configuration.set_value(
                                self.section, "conda-environment", conda_environment
                            )
                            self.configuration.set_value(self.section, "modules", "")
                            self.configuration.save()
                            print(
                                "Will use the conda environment "
                                f"'{conda_environment}'"
                            )
                            break
            if lammps_path is None:
                # Haven't found it. Check in the path.
                lammps_path = self.executables_in_path()
                if lammps_path is not None:
                    if self.options.yes or self.ask_yes_no(
                        "Found LAMMPS executables in the PATH at "
                        f"{lammps_path}\n"
                        "Use them?",
                        default="yes",
                    ):
                        self.configuration.set_value(
                            self.section, "installation", "path"
                        )
                        self.configuration.set_value(
                            self.section, "conda-environment", ""
                        )
                        self.configuration.set_value(self.section, "modules", "")
                        self.configuration.save()
                        print("Using the LAMMPS executables at {lammps_path}")

            if lammps_path is None:
                # Can't find LAMMPS
                print("Cannot find LAMMPS executables. You will need to install them.")
                if (
                    initial_installation is not None
                    and initial_installation != "not installed"
                ):
                    if self.options.yes or self.ask_yes_no(
                        "The configuration file indicates that LAMMPS "
                        "is installed, but it can't be found.\n"
                        "Fix the configuration file?",
                        default="yes",
                    ):
                        self.configuration.set_value(
                            self.section, "installation", "not installed"
                        )
                        self.configuration.set_value(self.section, "lammps-path", "")
                        self.configuration.set_value(
                            self.section, "conda-environment", ""
                        )
                        self.configuration.set_value(self.section, "modules", "")
                        self.configuration.save()
                        print(
                            "Since no LAMMPS executables were found, cleared "
                            "the configuration."
                        )
            else:
                print("The check completed successfully.")

    def check_configuration_file(self):
        """Checks that the lammps-step section is in the configuration file."""
        if not self.configuration.section_exists(self.section):
            # Get the text of the data
            path = Path(pkg_resources.resource_filename(__name__, "data/"))
            path = path / "configuration.txt"
            text = path.read_text()

            # Add it to the configuration file and write to disk.
            self.configuration.add_section(self.section, text)
            self.configuration.save()

    def have_executables(self, path):
        """Check whether the executables are found at the given path.

        Parameters
        ----------
        path : pathlib.Path
            The directory to check.

        Returns
        -------
        bool
            True if at least one of the LAMMPS executables is found.
        """
        for executable in self.executables:
            tmp_path = path / executable
            if tmp_path.exists():
                self.logger.debug(f"Found executables in {path}")
                return True
        self.logger.debug(f"Did not find executables in {path}")
        return False

    def executables_in_path(self):
        """Check whether the executables are found in the PATH.

        Returns
        -------
        pathlib.Path
            The path where the executables are, or None.
        """
        path = None
        for executable in self.executables:
            path = shutil.which(executable)
            if path is not None:
                path = Path(path).expanduser().resolve()
                break
        return path

    def install(self):
        """Install LAMMPS using a Conda environment."""
        if self.configuration.section_exists(self.section):
            # Get the values from the configuration
            data = self.configuration.get_values(self.section)
            if "installation" in data:
                initial_installation = data["installation"]
                if initial_installation == "path":
                    print("Using LAMMPS for the path given by the user.")
                    return
                elif initial_installation == "modules":
                    print("Using LAMMPS from the modules given by the user.")
                    return
        else:
            # Update the configuration file.
            self.check_configuration_file()

        print(
            f"Installing Conda environment '{self.environment}'. This "
            "may take a minute or two."
        )
        self.conda.create_environment(self.environment_file, name=self.environment)
        path = self.conda.path(self.environment) / "bin"
        self.configuration.set_value(self.section, "lammps-path", str(path))
        self.configuration.set_value(self.section, "installation", "conda")
        self.configuration.set_value(
            self.section, "conda-environment", self.environment
        )
        self.configuration.set_value(self.section, "modules", "")
        self.configuration.save()
        print("Done!\n")

    def show(self):
        """Show the current installation status."""
        self.logger.debug("Entering show")

        # See if LAMMPS is already registered in the configuration file
        if not self.configuration.section_exists(self.section):
            print(
                "There is no section in the configuration file for the "
                "LAMMPS step (lammps-step)."
            )
        data = self.configuration.get_values(self.section)

        # Keep track of where executables are
        serial = None
        mpi = None
        mpiexec = None

        # Is the path in the configuration file?
        if "lammps-path" in data:
            conf_path = Path(data["lammps-path"]).expanduser().resolve()
            if (conf_path / "lmp_serial").exists():
                serial = conf_path / "lmp_serial"
                serial_version = self.exe_version(serial)
            if (conf_path / "mpiexec").exists():
                mpiexec = conf_path / "mpiexec"
            else:
                mpiexec = shutil.which("mpiexec")
            if (conf_path / "lmp_mpi").exists():
                mpi = conf_path / "lmp_mpi"
                if mpiexec is None:
                    mpi_version = "unknown"
                else:
                    mpi_version = self.exe_version(mpi, mpiexec)

            extra = f"from path {conf_path}."
            if "installation" in data:
                installation = data["installation"]
                if installation == "conda":
                    if "conda-environment" in data and data["conda-environment"] != "":
                        extra = (
                            "from Conda environment " f"{data['conda-environment']}."
                        )
                    else:
                        extra = "from an unknown Conda environment."
                elif installation == "modules":
                    if "modules" in data and data["modules"] != "":
                        print(f"from module(s) {data['modules']}.")
                    else:
                        print("from unknown modules.")
                    return
                elif installation == "path":
                    print(f"from user-defined path {conf_path}.")
                    return

            if serial is not None:
                if mpi is not None:
                    if serial_version == mpi_version:
                        print(
                            "LAMMPS serial and mpi executables, version "
                            f"'{serial_version}'"
                        )
                    else:
                        print(
                            "LAMMPS serial executable, version "
                            f"'{serial_version}', and mpi executable, "
                            f"version {mpi_version}"
                        )
                    print(extra)
                else:
                    print(f"LAMMPS serial executable, version {serial_version}")
                    print(extra)
            elif mpi is not None:
                print(f"LAMMPS mpi executable, version {mpi_version}")
                print(extra)
            else:
                print("LAMMPS is not configured to run.")
        else:
            print("LAMMPS is not configured to run.")

        # Look in the PATH, but only record if not same as in the conf file
        tmp = shutil.which("lmp_serial")
        if tmp is not None:
            tmp = Path(tmp).expanduser().resolve()
            if serial is not None and serial != tmp:
                version = self.exe_version(tmp)
                print(
                    f"Another serial executable of LAMMPS (version {version}) "
                    "is in the PATH:\n"
                    f"    {tmp}"
                )
        tmp = shutil.which("lmp_mpi")
        if tmp is not None:
            tmp = Path(tmp).expanduser().resolve()
            if mpi is not None and mpi != tmp:
                if mpiexec is None:
                    version = "unknown"
                else:
                    version = self.exe_version(tmp, mpiexec)
                print(
                    f"Another mpi executable of LAMMPS (version {version}) "
                    "is in the PATH:\n"
                    f"    {tmp}"
                )

    def exe_version(self, path, mpiexec=None):
        """Get the version of the LAMMPS executable.

        Parameters
        ----------
        path : pathlib.Path
            Path to the executable.

        Returns
        -------
        str
            The version reported by LAMMPS, or 'unknown'.
        """
        if mpiexec is not None:
            try:
                result = subprocess.run(
                    [mpiexec, str(path), "-log", "none"],
                    stdin=subprocess.DEVNULL,
                    capture_output=True,
                    text=True,
                )
            except Exception:
                version = "unknown"
            else:
                version = "unknown"
                lines = result.stdout.splitlines()
                if len(lines) > 0:
                    line = lines[0]
                    if line[0:8] == "LAMMPS (":
                        version = line[8:].rstrip(")")
        else:
            try:
                result = subprocess.run(
                    [str(path), "-log", "none"],
                    stdin=subprocess.DEVNULL,
                    capture_output=True,
                    text=True,
                )
            except Exception:
                version = "unknown"
            else:
                version = "unknown"
                lines = result.stdout.splitlines()
                if len(lines) > 0:
                    line = lines[0]
                    if line[0:8] == "LAMMPS (":
                        version = line[8:].rstrip(")")

        return version



"""This file contains metadata to help describe the results of LAMMPS
calculations, etc.
"""

metadata = {}

"""Properties that LAMMPS produces, depending on the type of calculation.
"""

metadata["results"] = {
    "T": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "temperature",
        "dimensionality": "scalar",
        "property": "temperature#LAMMPS#{model}",
        "type": "float",
        "units": "K",
    },
    "T,stderr": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "stderr of temperature",
        "dimensionality": "scalar",
        "property": "temperature, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "K",
    },
    "T,tau": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "autocorrelation time of temperature",
        "dimensionality": "scalar",
        "property": "temperature, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "T,inefficiency": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "statistical inefficiency of temperature sampling",
        "dimensionality": "scalar",
        "property": "temperature, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "P": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "pressure",
        "dimensionality": "scalar",
        "property": "pressure#LAMMPS#{model}",
        "type": "float",
        "units": "atm",
    },
    "P,stderr": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "stderr of pressure",
        "dimensionality": "scalar",
        "property": "pressure, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "atm",
    },
    "P,tau": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "autocorrelation time of pressure",
        "dimensionality": "scalar",
        "property": "pressure, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "P,inefficiency": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "statistical inefficiency of pressure sampling",
        "dimensionality": "scalar",
        "property": "pressure, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "density": {
        "calculation": ["npt"],
        "description": "pressure",
        "dimensionality": "scalar",
        "property": "density#LAMMPS#{model}",
        "type": "float",
        "units": "g/ml",
    },
    "density,stderr": {
        "calculation": ["npt"],
        "description": "stderr of density",
        "dimensionality": "scalar",
        "property": "density, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "g/ml",
    },
    "density,tau": {
        "calculation": ["npt"],
        "description": "autocorrelation time of density",
        "dimensionality": "scalar",
        "property": "density, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "density,inefficiency": {
        "calculation": ["npt"],
        "description": "statistical inefficiency of density sampling",
        "dimensionality": "scalar",
        "property": "density, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "a": {
        "calculation": ["npt"],
        "description": "cell parameter 'a'",
        "dimensionality": "scalar",
        "property": "cell_a#LAMMPS#{model}",
        "type": "float",
        "units": "Å",
    },
    "a,stderr": {
        "calculation": ["npt"],
        "description": "stderr of cell 'a'",
        "dimensionality": "scalar",
        "property": "cell_a, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "Å",
    },
    "a,tau": {
        "calculation": ["npt"],
        "description": "autocorrelation time of cell 'a'",
        "dimensionality": "scalar",
        "property": "cell_a, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "a,inefficiency": {
        "calculation": ["npt"],
        "description": "statistical inefficiency of cell 'a' sampling",
        "dimensionality": "scalar",
        "property": "cell_a, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "b": {
        "calculation": ["npt"],
        "description": "cell parameter 'b'",
        "dimensionality": "scalar",
        "property": "cell_b#LAMMPS#{model}",
        "type": "float",
        "units": "Å",
    },
    "b,stderr": {
        "calculation": ["npt"],
        "description": "stderr of cell 'b'",
        "dimensionality": "scalar",
        "property": "cell_b, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "Å",
    },
    "b,tau": {
        "calculation": ["npt"],
        "description": "autocorrelation time of cell 'b'",
        "dimensionality": "scalar",
        "property": "cell_b, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "b,inefficiency": {
        "calculation": ["npt"],
        "description": "statistical inefficiency of cell 'b' sampling",
        "dimensionality": "scalar",
        "property": "cell_b, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "c": {
        "calculation": ["npt"],
        "description": "cell parameter 'c'",
        "dimensionality": "scalar",
        "property": "cell_c#LAMMPS#{model}",
        "type": "float",
        "units": "Å",
    },
    "c,stderr": {
        "calculation": ["npt"],
        "description": "stderr of cell 'c'",
        "dimensionality": "scalar",
        "property": "cell_c, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "Å",
    },
    "c,tau": {
        "calculation": ["npt"],
        "description": "autocorrelation time of cell 'c'",
        "dimensionality": "scalar",
        "property": "cell_c, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "c,inefficiency": {
        "calculation": ["npt"],
        "description": "statistical inefficiency of cell 'c' sampling",
        "dimensionality": "scalar",
        "property": "cell_c, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "Etot": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "total energy",
        "dimensionality": "scalar",
        "property": "total energy#LAMMPS#{model}",
        "type": "float",
        "units": "kcal/mol",
    },
    "Etot,stderr": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "stderr of total energy",
        "dimensionality": "scalar",
        "property": "total energy, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "kcal/mol",
    },
    "Etot,tau": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "autocorrelation time of total energy",
        "dimensionality": "scalar",
        "property": "total energy, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "Etot,inefficiency": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "statistical inefficiency of total energy sampling",
        "dimensionality": "scalar",
        "property": "total energy, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "Eke": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "kinetic energy",
        "dimensionality": "scalar",
        "property": "kinetic energy#LAMMPS#{model}",
        "type": "float",
        "units": "kcal/mol",
    },
    "Eke,stderr": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "stderr of kinetic energy",
        "dimensionality": "scalar",
        "property": "kinetic energy, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "kcal/mol",
    },
    "Eke,tau": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "autocorrelation time of kinetic energy",
        "dimensionality": "scalar",
        "property": "kinetic energy, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "Eke,inefficiency": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "statistical inefficiency of kinetic energy sampling",
        "dimensionality": "scalar",
        "property": "kinetic energy, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "Epe": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "potential energy",
        "dimensionality": "scalar",
        "property": "potential energy#LAMMPS#{model}",
        "type": "float",
        "units": "kcal/mol",
    },
    "Epe,stderr": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "stderr of potential energy",
        "dimensionality": "scalar",
        "property": "potential energy, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "kcal/mol",
    },
    "Epe,tau": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "autocorrelation time of potential energy",
        "dimensionality": "scalar",
        "property": "potential energy, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "Epe,inefficiency": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "statistical inefficiency of potential energy sampling",
        "dimensionality": "scalar",
        "property": "potential energy, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "Epair": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "nonbonded (vdW & electrostatic) energy",
        "dimensionality": "scalar",
        "type": "float",
        "units": "kcal/mol",
    },
    "Epair,stderr": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "stderr of nonbond energy",
        "dimensionality": "scalar",
        "type": "float",
        "units": "kcal/mol",
    },
    "Epair,tau": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "autocorrelation time of nonbond energy",
        "dimensionality": "scalar",
        "type": "float",
        "units": "fs",
    },
    "Epair,inefficiency": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "statistical inefficiency of nonbond energy sampling",
        "dimensionality": "scalar",
        "type": "float",
        "units": "",
    },
}
